@@ -13,6 +13,7 @@ use crate::ast::UnaryOperator;
 use crate::ast::AST;
 use santiago::grammar::Associativity;
 use santiago::grammar::Grammar;
+use santiago::lexer::Position;
 use std::collections::LinkedList;
 
 pub fn grammar() -> Grammar<AST> {
@@ -20,59 +21,77 @@ pub fn grammar() -> Grammar<AST> {
         "expr" => rules "expr_function";
 
         "expr_function" => rules "ID" ":" "expr_function"
-            => |mut asts| AST::Function {
-                argument: Some(match asts.swap_remove(0) {
-                    AST::__Lexeme(lexeme) => lexeme.raw.clone(),
-                    _ => unreachable!(),
-                }),
-                arguments: FunctionArguments {
-                    arguments: LinkedList::new(),
-                    ellipsis: false,
+            => |mut asts| match asts.swap_remove(0) {
+                AST::__Lexeme(lexeme) => AST::Function {
+                    argument: Some(lexeme.raw.clone()),
+                    arguments: FunctionArguments {
+                        arguments: LinkedList::new(),
+                        ellipsis: false,
+                    },
+                    definition: Box::new(asts.swap_remove(0)),
+                    position: lexeme.position.clone(),
                 },
-                definition: Box::new(asts.swap_remove(0)),
+                _ => unreachable!(),
             };
         "expr_function" => rules "{" "formals" "}" ":" "expr_function"
-            => |mut asts| AST::Function {
-                argument: None,
-                arguments: match asts.swap_remove(1) {
-                    AST::__FunctionArguments(function_arguments) => function_arguments,
-                    _ => unreachable!(),
+            => |mut asts| match asts.swap_remove(0) {
+                AST::__Lexeme(lexeme) => AST::Function {
+                    argument: None,
+                    arguments: match asts.swap_remove(1) {
+                        AST::__FunctionArguments(function_arguments) => function_arguments,
+                        _ => unreachable!(),
+                    },
+                    definition: Box::new(asts.swap_remove(0)),
+                    position: lexeme.position.clone(),
                 },
-                definition: Box::new(asts.swap_remove(1)),
+                _ => unreachable!(),
             };
         "expr_function" => rules "{" "formals" "}" "@" "ID" ":" "expr_function"
-            => |mut asts| AST::Function {
-                argument: Some(match asts.swap_remove(4) {
-                    AST::__Lexeme(lexeme) => lexeme.raw.clone(),
-                    _ => unreachable!(),
-                }),
-                arguments: match asts.swap_remove(1) {
-                    AST::__FunctionArguments(function_arguments) => function_arguments,
-                    _ => unreachable!(),
+            => |mut asts| match asts.swap_remove(0) {
+                AST::__Lexeme(lexeme) => AST::Function {
+                    argument: Some(match asts.swap_remove(4) {
+                        AST::__Lexeme(lexeme) => lexeme.raw.clone(),
+                        _ => unreachable!(),
+                    }),
+                    arguments: match asts.swap_remove(1) {
+                        AST::__FunctionArguments(function_arguments) => function_arguments,
+                        _ => unreachable!(),
+                    },
+                    definition: Box::new(asts.swap_remove(0)),
+                    position: lexeme.position.clone(),
                 },
-                definition: Box::new(asts.swap_remove(4)),
+                _ => unreachable!(),
             };
         "expr_function" => rules "ID" "@" "{" "formals" "}" ":" "expr_function"
-            => |mut asts| AST::Function {
-                argument: Some(match asts.swap_remove(0) {
-                    AST::__Lexeme(lexeme) => lexeme.raw.clone(),
-                    _ => unreachable!(),
-                }),
-                arguments: match asts.swap_remove(3) {
-                    AST::__FunctionArguments(function_arguments) => function_arguments,
-                    _ => unreachable!(),
+            => |mut asts| match asts.swap_remove(0) {
+                AST::__Lexeme(lexeme) => AST::Function {
+                    argument: Some(lexeme.raw.clone()),
+                    arguments: match asts.swap_remove(3) {
+                        AST::__FunctionArguments(function_arguments) => function_arguments,
+                        _ => unreachable!(),
+                    },
+                    definition: Box::new(asts.swap_remove(0)),
+                    position: lexeme.position.clone(),
                 },
-                definition: Box::new(asts.swap_remove(0)),
+                _ => unreachable!(),
             };
         "expr_function" => rules "ASSERT" "expr" ";" "expr_function"
             => |mut asts| AST::Assert {
                 expression: Box::new(asts.swap_remove(1)),
                 target: Box::new(asts.swap_remove(1)),
+                position: match asts.swap_remove(0) {
+                    AST::__Lexeme(lexeme) => lexeme.position.clone(),
+                    _ => unreachable!(),
+                },
             };
         "expr_function" => rules "WITH" "expr" ";" "expr_function"
             => |mut asts| AST::With {
                 expression: Box::new(asts.swap_remove(1)),
                 target: Box::new(asts.swap_remove(1)),
+                position: match asts.swap_remove(0) {
+                    AST::__Lexeme(lexeme) => lexeme.position.clone(),
+                    _ => unreachable!(),
+                },
             };
         "expr_function" => rules "LET" "binds" "IN" "expr_function"
             => |mut asts| {
@@ -84,6 +103,10 @@ pub fn grammar() -> Grammar<AST> {
                         _ => unreachable!(),
                     },
                     target: Box::new(asts.swap_remove(1)),
+                    position: match asts.swap_remove(0) {
+                        AST::__Lexeme(lexeme) => lexeme.position.clone(),
+                        _ => unreachable!(),
+                    },
                 }
             };
         "expr_function" => rules "expr_if";
@@ -93,6 +116,10 @@ pub fn grammar() -> Grammar<AST> {
                 predicate: Box::new(asts.swap_remove(1)),
                 then: Box::new(asts.swap_remove(3)),
                 else_: Box::new(asts.swap_remove(1)),
+                position: match asts.swap_remove(0) {
+                    AST::__Lexeme(lexeme) => lexeme.position.clone(),
+                    _ => unreachable!(),
+                },
             };
         "expr_if" => rules "expr_op";
 
@@ -100,11 +127,19 @@ pub fn grammar() -> Grammar<AST> {
             => |mut asts| AST::UnaryOperation {
                 operator: UnaryOperator::Not,
                 operand: Box::new(asts.swap_remove(1)),
+                position: match asts.swap_remove(0) {
+                    AST::__Lexeme(lexeme) => lexeme.position.clone(),
+                    _ => unreachable!(),
+                },
             };
         "expr_op" => rules "NEGATE" "expr_op"
             => |mut asts| AST::UnaryOperation {
                 operator: UnaryOperator::Negate,
                 operand: Box::new(asts.swap_remove(1)),
+                position: match asts.swap_remove(0) {
+                    AST::__Lexeme(lexeme) => lexeme.position.clone(),
+                    _ => unreachable!(),
+                },
             };
         "expr_op" => rules "expr_op" "EQ" "expr_op"
             => |mut asts| AST::BinaryOperation {
@@ -113,6 +148,10 @@ pub fn grammar() -> Grammar<AST> {
                     asts.swap_remove(0),
                     asts.swap_remove(0),
                 ],
+                position: match asts.swap_remove(0) {
+                    AST::__Lexeme(lexeme) => lexeme.position.clone(),
+                    _ => unreachable!(),
+                },
             };
         "expr_op" => rules "expr_op" "NEQ" "expr_op"
             => |mut asts| AST::BinaryOperation {
@@ -121,6 +160,10 @@ pub fn grammar() -> Grammar<AST> {
                     asts.swap_remove(0),
                     asts.swap_remove(0),
                 ],
+                position: match asts.swap_remove(0) {
+                    AST::__Lexeme(lexeme) => lexeme.position.clone(),
+                    _ => unreachable!(),
+                },
             };
         "expr_op" => rules "expr_op" "<" "expr_op"
             => |mut asts| AST::BinaryOperation {
@@ -129,6 +172,10 @@ pub fn grammar() -> Grammar<AST> {
                     asts.swap_remove(0),
                     asts.swap_remove(0),
                 ],
+                position: match asts.swap_remove(0) {
+                    AST::__Lexeme(lexeme) => lexeme.position.clone(),
+                    _ => unreachable!(),
+                },
             };
         "expr_op" => rules "expr_op" "LEQ" "expr_op"
             => |mut asts| AST::BinaryOperation {
@@ -137,6 +184,10 @@ pub fn grammar() -> Grammar<AST> {
                     asts.swap_remove(0),
                     asts.swap_remove(0),
                 ],
+                position: match asts.swap_remove(0) {
+                    AST::__Lexeme(lexeme) => lexeme.position.clone(),
+                    _ => unreachable!(),
+                },
             };
         "expr_op" => rules "expr_op" ">" "expr_op"
             => |mut asts| AST::BinaryOperation {
@@ -145,6 +196,10 @@ pub fn grammar() -> Grammar<AST> {
                     asts.swap_remove(0),
                     asts.swap_remove(0),
                 ],
+                position: match asts.swap_remove(0) {
+                    AST::__Lexeme(lexeme) => lexeme.position.clone(),
+                    _ => unreachable!(),
+                },
             };
         "expr_op" => rules "expr_op" "GEQ" "expr_op"
             => |mut asts| AST::BinaryOperation {
@@ -153,6 +208,10 @@ pub fn grammar() -> Grammar<AST> {
                     asts.swap_remove(0),
                     asts.swap_remove(0),
                 ],
+                position: match asts.swap_remove(0) {
+                    AST::__Lexeme(lexeme) => lexeme.position.clone(),
+                    _ => unreachable!(),
+                },
             };
         "expr_op" => rules "expr_op" "AND" "expr_op"
             => |mut asts| AST::BinaryOperation {
@@ -161,6 +220,10 @@ pub fn grammar() -> Grammar<AST> {
                     asts.swap_remove(0),
                     asts.swap_remove(0),
                 ],
+                position: match asts.swap_remove(0) {
+                    AST::__Lexeme(lexeme) => lexeme.position.clone(),
+                    _ => unreachable!(),
+                },
             };
         "expr_op" => rules "expr_op" "OR" "expr_op"
             => |mut asts| AST::BinaryOperation {
@@ -169,6 +232,10 @@ pub fn grammar() -> Grammar<AST> {
                     asts.swap_remove(0),
                     asts.swap_remove(0),
                 ],
+                position: match asts.swap_remove(0) {
+                    AST::__Lexeme(lexeme) => lexeme.position.clone(),
+                    _ => unreachable!(),
+                },
             };
         "expr_op" => rules "expr_op" "IMPL" "expr_op"
             => |mut asts| AST::BinaryOperation {
@@ -177,6 +244,10 @@ pub fn grammar() -> Grammar<AST> {
                     asts.swap_remove(0),
                     asts.swap_remove(0),
                 ],
+                position: match asts.swap_remove(0) {
+                    AST::__Lexeme(lexeme) => lexeme.position.clone(),
+                    _ => unreachable!(),
+                },
             };
         "expr_op" => rules "expr_op" "UPDATE" "expr_op"
             => |mut asts| AST::BinaryOperation {
@@ -185,14 +256,22 @@ pub fn grammar() -> Grammar<AST> {
                     asts.swap_remove(0),
                     asts.swap_remove(0),
                 ],
+                position: match asts.swap_remove(0) {
+                    AST::__Lexeme(lexeme) => lexeme.position.clone(),
+                    _ => unreachable!(),
+                },
             };
         "expr_op" => rules "expr_op" "?" "attrpath"
-            => |mut asts| AST::BinaryOperation {
-                operator: BinaryOperator::HasAttribute,
-                operands: vec![
-                    asts.swap_remove(0),
-                    asts.swap_remove(0),
-                ],
+            => |mut asts| AST::HasProperty {
+                expression: Box::new(asts.swap_remove(0)),
+                attribute_path: match asts.swap_remove(0) {
+                    AST::__AttributePath(attribute_path) => attribute_path,
+                    _ => unreachable!(),
+                },
+                position: match asts.swap_remove(0) {
+                    AST::__Lexeme(lexeme) => lexeme.position.clone(),
+                    _ => unreachable!(),
+                },
             };
         "expr_op" => rules "expr_op" "+" "expr_op"
             => |mut asts| AST::BinaryOperation {
@@ -201,6 +280,10 @@ pub fn grammar() -> Grammar<AST> {
                     asts.swap_remove(0),
                     asts.swap_remove(0),
                 ],
+                position: match asts.swap_remove(0) {
+                    AST::__Lexeme(lexeme) => lexeme.position.clone(),
+                    _ => unreachable!(),
+                },
             };
         "expr_op" => rules "expr_op" "-" "expr_op"
             => |mut asts| AST::BinaryOperation {
@@ -209,6 +292,10 @@ pub fn grammar() -> Grammar<AST> {
                     asts.swap_remove(0),
                     asts.swap_remove(0),
                 ],
+                position: match asts.swap_remove(0) {
+                    AST::__Lexeme(lexeme) => lexeme.position.clone(),
+                    _ => unreachable!(),
+                },
             };
         "expr_op" => rules "expr_op" "*" "expr_op"
             => |mut asts| AST::BinaryOperation {
@@ -217,6 +304,10 @@ pub fn grammar() -> Grammar<AST> {
                     asts.swap_remove(0),
                     asts.swap_remove(0),
                 ],
+                position: match asts.swap_remove(0) {
+                    AST::__Lexeme(lexeme) => lexeme.position.clone(),
+                    _ => unreachable!(),
+                },
             };
         "expr_op" => rules "expr_op" "/" "expr_op"
             => |mut asts| AST::BinaryOperation {
@@ -225,6 +316,10 @@ pub fn grammar() -> Grammar<AST> {
                     asts.swap_remove(0),
                     asts.swap_remove(0),
                 ],
+                position: match asts.swap_remove(0) {
+                    AST::__Lexeme(lexeme) => lexeme.position.clone(),
+                    _ => unreachable!(),
+                },
             };
         "expr_op" => rules "expr_op" "CONCAT" "expr_op"
             => |mut asts| AST::BinaryOperation {
@@ -233,6 +328,10 @@ pub fn grammar() -> Grammar<AST> {
                     asts.swap_remove(0),
                     asts.swap_remove(0),
                 ],
+                position: match asts.swap_remove(0) {
+                    AST::__Lexeme(lexeme) => lexeme.position.clone(),
+                    _ => unreachable!(),
+                },
             };
         "expr_op" => rules "expr_app";
 
@@ -273,7 +372,15 @@ pub fn grammar() -> Grammar<AST> {
         "expr_select" => rules "expr_simple" "OR_KW"
             => |mut asts| AST::FunctionApplication {
                 function: Box::new(asts.swap_remove(0)),
-                arguments: LinkedList::from([asts.swap_remove(0)]),
+                arguments: LinkedList::from([
+                    match asts.swap_remove(0) {
+                        AST::__Lexeme(lexeme) => AST::Variable {
+                            identifier: lexeme.raw.clone(),
+                            position: lexeme.position.clone(),
+                        },
+                        _ => unreachable!(),
+                    }
+                ]),
             };
         "expr_select" => rules "expr_simple";
 
@@ -348,7 +455,10 @@ pub fn grammar() -> Grammar<AST> {
             };
         "expr_simple" => rules "URI"
             => |mut asts| match asts.swap_remove(0) {
-                AST::__Lexeme(lexeme) => AST::Uri(lexeme),
+                AST::__Lexeme(lexeme) => AST::Uri {
+                    uri: lexeme.raw.clone(),
+                    position: lexeme.position.clone(),
+                },
                 _ => unreachable!(),
             };
         "expr_simple" => rules "(" "expr" ")"
@@ -363,6 +473,10 @@ pub fn grammar() -> Grammar<AST> {
                         AST::__Bindings(bindings) => bindings,
                         _ => unreachable!(),
                     },
+                    position: match asts.swap_remove(0) {
+                        AST::__Lexeme(lexeme) => lexeme.position.clone(),
+                        _ => unreachable!(),
+                    },
                     recursive: true,
                 }
             };
@@ -375,11 +489,29 @@ pub fn grammar() -> Grammar<AST> {
                         AST::__Bindings(bindings) => bindings,
                         _ => unreachable!(),
                     },
+                    position: match asts.swap_remove(0) {
+                        AST::__Lexeme(lexeme) => lexeme.position.clone(),
+                        _ => unreachable!(),
+                    },
                     recursive: false,
                 }
             };
         "expr_simple" => rules "[" "expr_list" "]"
-            => |mut asts| asts.swap_remove(1);
+            => |mut asts| {
+                let mut expr_list = asts.swap_remove(1);
+
+                match &mut expr_list {
+                    AST::List { position, .. } => {
+                        *position = match asts.swap_remove(0) {
+                            AST::__Lexeme(lexeme) => lexeme.position.clone(),
+                            _ => unreachable!(),
+                        };
+                    },
+                    _ => unreachable!(),
+                }
+
+                expr_list
+            };
 
         "string_parts" => rules "STR"
             => |mut asts| AST::String {
@@ -651,19 +783,19 @@ pub fn grammar() -> Grammar<AST> {
         "attrpath" => rules "attr"
             => |mut asts| AST::__AttributePath(AttributePath {
                 attributes: LinkedList::from([
-                (match asts.swap_remove(0) {
-                    AST::__Attribute(attribute) => attribute,
-                    _ => unreachable!(),
-                })
+                    (match asts.swap_remove(0) {
+                        AST::__Attribute(attribute) => attribute,
+                        _ => unreachable!(),
+                    })
                 ]),
             });
         "attrpath" => rules "string_attr"
             => |mut asts| AST::__AttributePath(AttributePath {
                 attributes: LinkedList::from([
-                (match asts.swap_remove(0) {
-                    AST::__Attribute(attribute) => attribute,
-                    _ => unreachable!(),
-                })
+                    (match asts.swap_remove(0) {
+                        AST::__Attribute(attribute) => attribute,
+                        _ => unreachable!(),
+                    })
                 ]),
             });
         "attr" => rules "ID"
@@ -701,7 +833,7 @@ pub fn grammar() -> Grammar<AST> {
                 let mut expr_list = asts.swap_remove(0);
 
                 match &mut expr_list {
-                    AST::List { elements } => {
+                    AST::List { elements, .. } => {
                         elements.push_back(asts.swap_remove(0));
                     }
                     _ => unreachable!(),
@@ -710,7 +842,13 @@ pub fn grammar() -> Grammar<AST> {
                 expr_list
             };
         "expr_list" => empty
-            => |_| AST::List { elements: LinkedList::new() };
+            => |_| AST::List {
+                elements: LinkedList::new(),
+                position: Position {
+                    line: usize::MAX,
+                    column: usize::MAX,
+                },
+            };
 
         "formals" => rules "formal" "," "formals"
             => |mut asts| {
